"""Jedi Language Server

Creates the language server constant and wraps "features" with it.

Official language server spec:
    https://microsoft.github.io/language-server-protocol/specification
"""

from typing import Dict  # pylint: disable=unused-import
from typing import List, Optional

from pygls.features import (
    COMPLETION,
    DEFINITION,
    DOCUMENT_SYMBOL,
    HOVER,
    REFERENCES,
    RENAME,
)
from pygls.server import LanguageServer
from pygls.types import (
    CompletionItem,
    CompletionList,
    CompletionParams,
    DocumentSymbolParams,
    Hover,
    Location,
    RenameParams,
    SymbolInformation,
    TextDocumentPositionParams,
    TextEdit,
    WorkspaceEdit,
)

from .server_utils import (
    get_jedi_script,
    get_location_from_name,
    get_symbol_information_from_name,
)
from .type_map import get_lsp_completion_type

SERVER = LanguageServer()


@SERVER.feature(COMPLETION, triggerCharacters=["."])
def lsp_completion(server: LanguageServer, params: CompletionParams):
    """Returns completion items."""
    script = get_jedi_script(server, params.textDocument.uri)
    jedi_completions = script.complete(
        line=params.position.line + 1, column=params.position.character,
    )
    return CompletionList(
        is_incomplete=False,
        items=[
            CompletionItem(
                label=completion.name,
                kind=get_lsp_completion_type(completion.type),
                detail=completion.description,
                documentation=completion.docstring(),
                insert_text=completion.name,
            )
            for completion in jedi_completions
        ],
    )


@SERVER.feature(DEFINITION)
def lsp_definition(
    server: LanguageServer, params: TextDocumentPositionParams
) -> List[Location]:
    """Support Goto Definition"""
    script = get_jedi_script(server, params.textDocument.uri)
    names = script.goto(
        line=params.position.line + 1,
        column=params.position.character,
        follow_imports=True,
        follow_builtin_imports=True,
    )
    return [get_location_from_name(name) for name in names]


@SERVER.feature(HOVER)
def lsp_hover(
    server: LanguageServer, params: TextDocumentPositionParams
) -> Hover:
    """Support the hover feature"""
    script = get_jedi_script(server, params.textDocument.uri)
    names = script.infer(
        line=params.position.line + 1, column=params.position.character,
    )
    return Hover(
        contents=(
            names[0].docstring() if names else "No docstring definition found."
        )
    )


@SERVER.feature(REFERENCES)
def lsp_references(
    server: LanguageServer, params: TextDocumentPositionParams
) -> List[Location]:
    """Obtain all references to document"""
    script = get_jedi_script(server, params.textDocument.uri)
    try:
        names = script.get_references(
            line=params.position.line + 1, column=params.position.character,
        )
    except Exception:  # pylint: disable=broad-except
        return []
    return [get_location_from_name(name) for name in names]


@SERVER.feature(RENAME)
def lsp_rename(
    server: LanguageServer, params: RenameParams
) -> Optional[WorkspaceEdit]:
    """Rename a symbol across a workspace"""
    script = get_jedi_script(server, params.textDocument.uri)
    try:
        names = script.get_references(
            line=params.position.line + 1, column=params.position.character,
        )
    except Exception:  # pylint: disable=broad-except
        return None
    locations = [get_location_from_name(name) for name in names]
    if not locations:
        return None
    changes = {}  # type: Dict[str, List[TextEdit]]
    for location in locations:
        text_edit = TextEdit(location.range, new_text=params.newName)
        if location.uri not in changes:
            changes[location.uri] = [text_edit]
        else:
            changes[location.uri].append(text_edit)
    return WorkspaceEdit(changes=changes)


@SERVER.feature(DOCUMENT_SYMBOL)
def lsp_document_symbol(
    server: LanguageServer, params: DocumentSymbolParams
) -> List[SymbolInformation]:
    """Document Python document symbols"""
    script = get_jedi_script(server, params.textDocument.uri)
    try:
        names = script.get_names(
            line=params.position.line + 1, column=params.position.character,
        )
    except Exception:  # pylint: disable=broad-except
        return []
    return [get_symbol_information_from_name(name) for name in names]



"""Utility functions used by the language server"""

from typing import Optional

from jedi import Project, Script
from jedi.api.classes import Name
from jedi.api.environment import get_cached_default_environment
from pygls.server import LanguageServer
from pygls.types import Location, Position, Range, SymbolInformation
from pygls.uris import from_fs_path

from .type_map import get_lsp_symbol_type


def get_jedi_script(server: LanguageServer, doc_uri: str) -> Script:
    """Simplifies getting jedi Script

    :param doc_uri: the uri for the LSP text document. Obtained through
    params.textDocument.uri

    NOTE:
        * jedi is 1-indexed for lines and 0-indexed for columns
        * LSP is 0-indexed for lines and 0-indexed for columns
        * Therefore, add 1 to LSP's request for the line
    """
    workspace = server.workspace
    text_doc = workspace.get_document(doc_uri)
    project = Project(
        path=workspace.root_path,
        smart_sys_path=True,
        load_unsafe_extensions=False,
    )
    return Script(
        code=text_doc.source,
        path=text_doc.path,
        project=project,
        environment=get_cached_default_environment(),
    )


def get_location_from_name(name: Name) -> Location:
    """Get LSP location from Jedi definition

    NOTE:
        * jedi is 1-indexed for lines and 0-indexed for columns
        * LSP is 0-indexed for lines and 0-indexed for columns
        * Therefore, subtract 1 from Jedi's definition line
    """
    return Location(
        uri=from_fs_path(name.module_path),
        range=Range(
            start=Position(line=name.line - 1, character=name.column),
            end=Position(
                line=name.line - 1, character=name.column + len(name.name),
            ),
        ),
    )


def get_symbol_information_from_name(name: Name,) -> SymbolInformation:
    """Get LSP SymbolInformation from Jedi definition"""
    return SymbolInformation(
        name=name.name,
        kind=get_lsp_symbol_type(name.type),
        location=get_location_from_name(name),
        container_name=get_jedi_parent_name(name),
    )


def get_jedi_parent_name(name: Name) -> Optional[str]:
    """Retrieve the parent name from Jedi

    Error prone Jedi calls are wrapped in try/except to avoid
    """
    try:
        parent = name.parent()
    except Exception:  # pylint: disable=broad-except
        return None
    return parent.name if parent and parent.parent() else None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jedi_language_server']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0', 'jedi>=0.17.0', 'pygls>=0.8.1']

entry_points = \
{'console_scripts': ['jedi-language-server = jedi_language_server.cli:cli']}

setup_kwargs = {
    'name': 'jedi-language-server',
    'version': '0.5.0',
    'description': 'A language server for Jedi!',
    'long_description': '# jedi-language-server\n\n[![image-version](https://img.shields.io/pypi/v/jedi-language-server.svg)](https://python.org/pypi/jedi-language-server)\n[![image-license](https://img.shields.io/pypi/l/jedi-language-server.svg)](https://python.org/pypi/jedi-language-server)\n[![image-python-versions](https://img.shields.io/pypi/pyversions/jedi-language-server.svg)](https://python.org/pypi/jedi-language-server)\n\nA [Language Server](https://microsoft.github.io/language-server-protocol/) for the latest version(s) of [Jedi](https://jedi.readthedocs.io/en/latest/).\n\n**Note:** this tool is actively used by its primary author. He\'s happy to review pull requests / respond to issues you may discover.\n\n## Installation\n\nFrom your command line (bash / zsh), run:\n\n```bash\npip install -U jedi jedi-language-server\n```\n\n`-U` ensures that you\'re pulling the latest version from pypi.\n\n## Overview\n\njedi-language-server aims to support all of Jedi\'s capabilities and expose them through the Language Server Protocol. It currently supports the following Language Server requests:\n\n- [textDocument/completion](https://microsoft.github.io/language-server-protocol/specifications/specification-current/#textDocument_completion)\n- [textDocument/definition](https://microsoft.github.io/language-server-protocol/specifications/specification-current/#textDocument_definition)\n- [textDocument/documentSymbol](https://microsoft.github.io/language-server-protocol/specifications/specification-current/#textDocument_documentSymbol)\n- [textDocument/hover](https://microsoft.github.io/language-server-protocol/specifications/specification-current/#textDocument_hover)\n- [textDocument/references](https://microsoft.github.io/language-server-protocol/specifications/specification-current/#textDocument_references)\n- [textDocument/rename](https://microsoft.github.io/language-server-protocol/specifications/specification-current/#textDocument_rename)\n\nThese language server requests are not currently configurable by the user, but we expect to relax this constraint in a future release.\n\n## Usage\n\nThe following instructions show how to use jedi-language-server with your development tooling. The instructions assume you have already installed jedi-language-server.\n\n### Command line (bash / zsh)\n\nAt your terminal prompt:\n\n```bash\njedi-language-server\n```\n\njedi-language-server currently works only over IO. This may change in the future.\n\n### Neovim\n\nConfigure jedi-language-server with [coc.nvim](https://github.com/neoclide/coc.nvim/wiki/Language-servers#register-custom-language-servers). For diagnostics, we recommend installing and using the latest version of [coc-diagnostic](https://github.com/iamcco/coc-diagnostic) with [pylint](https://github.com/PyCQA/pylint).\n\n~/.config/nvim/coc-settings.json:\n\n```json\n"languageserver": {\n  "jls": {\n    "command": "jedi-language-server",\n    "args": [],\n    "filetypes": ["python"]\n  }\n},\n"diagnostic-languageserver.filetypes": {\n  "python": "pylint"\n},\n"diagnostic-languageserver.linters": {\n  "pylint": {\n    "sourceName": "pylint",\n    "command": "pylint",\n    "args": [\n      "--output-format",\n      "text",\n      "--score",\n      "no",\n      "--msg-template",\n      "\'{line}:{column}:{category}:{msg} ({msg_id}:{symbol})\'",\n      "%file"\n    ],\n    "formatPattern": [\n      "^(\\\\d+?):(\\\\d+?):([a-z]+?):(.*)$",\n      {\n        "line": 1,\n        "column": 2,\n        "security": 3,\n        "message": 4\n      }\n    ],\n    "securities": {\n      "informational": "hint",\n      "refactor": "info",\n      "convention": "info",\n      "warning": "warning",\n      "error": "error",\n      "fatal": "error"\n    },\n    "offsetColumn": 1,\n    "formatLines": 1\n  }\n}\n```\n\n## Local Development\n\nTo build and run this project from source:\n\n### Dependencies\n\nInstall the following tools manually:\n\n- [Poetry](https://github.com/sdispater/poetry#installation)\n- [GNU Make](https://www.gnu.org/software/make/)\n\n#### Recommended\n\n- [asdf](https://github.com/asdf-vm/asdf)\n\n### Get source code\n\n[Fork](https://help.github.com/en/github/getting-started-with-github/fork-a-repo) this repository and clone the fork to your development machine:\n\n```bash\ngit clone https://github.com/<YOUR-USERNAME>/jedi-language-server\ncd jedi-language-server\n```\n\n### Set up development environment\n\n```bash\nmake setup\n```\n\n### Run tests\n\n```bash\nmake test\n```\n\n## Inspiration\n\nPalantir\'s [python-language-server](https://github.com/palantir/python-language-server) inspired this project. Unlike python-language-server, jedi-language-server:\n\n- Uses `pygls` instead of creating its own low-level Language Server Protocol bindings\n- Supports one powerful 3rd party library: Jedi. By only supporting Jedi, we can focus on supporting all Jedi features without exposing ourselves to too many broken 3rd party dependencies (I\'m looking at you, [rope](https://github.com/python-rope/rope)).\n- Is supremely simple because of its scope constraints. Leave complexity to the Jedi [master](https://github.com/davidhalter). If the force is strong with you, please submit a PR!\n\n## Written by\n\nSamuel Roeca _samuel.roeca@gmail.com_\n',
    'author': 'Sam Roeca',
    'author_email': 'samuel.roeca@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pappasam/jedi-language-server',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""Jedi Language Server command line interface"""

import click

from .server import SERVER


@click.command()
@click.version_option()
def cli() -> None:
    """Jedi language server

    Examples:

        Run from stdio : jedi-language-server
    """
    SERVER.start_io()
